@@ -1,18 +1,31 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from sql import Null
-from sql.operators import Concat
+from functools import wraps
 
 from trytond.model import Workflow, ModelView, fields
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
-from trytond import backend
 
 __all__ = ['ShipmentOut', 'ShipmentOutReturn', 'Move']
 
 
-class ShipmentOut:
-    __metaclass__ = PoolMeta
+def process_sale(moves_field):
+    def _process_sale(func):
+        @wraps(func)
+        def wrapper(cls, shipments):
+            pool = Pool()
+            Sale = pool.get('sale.sale')
+            with Transaction().set_context(_check_access=False):
+                sales = set(m.sale for s in cls.browse(shipments)
+                    for m in getattr(s, moves_field) if m.sale)
+            func(cls, shipments)
+            if sales:
+                Sale.__queue__.process(sales)
+        return wrapper
+    return _process_sale
+
+
+class ShipmentOut(metaclass=PoolMeta):
     __name__ = 'stock.shipment.out'
 
     @classmethod
@@ -23,31 +36,6 @@ class ShipmentOut:
                     'by a sale.'),
                 })
 
-    @classmethod
-    def write(cls, *args):
-        pool = Pool()
-        Sale = pool.get('sale.sale')
-        SaleLine = pool.get('sale.line')
-
-        super(ShipmentOut, cls).write(*args)
-
-        actions = iter(args)
-        for shipments, values in zip(actions, actions):
-            if values.get('state') not in ('done', 'cancel'):
-                continue
-            sales = []
-            move_ids = []
-            for shipment in shipments:
-                move_ids.extend([x.id for x in shipment.outgoing_moves])
-
-            with Transaction().set_context(_check_access=False):
-                sale_lines = SaleLine.search([
-                        ('moves', 'in', move_ids),
-                        ])
-                if sale_lines:
-                    sales = list(set(l.sale for l in sale_lines))
-                    Sale.process(sales)
-
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
@@ -61,9 +49,22 @@ class ShipmentOut:
 
         return super(ShipmentOut, cls).draft(shipments)
 
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('done')
+    @process_sale('outgoing_moves')
+    def done(cls, shipments):
+        super(ShipmentOut, cls).done(shipments)
 
-class ShipmentOutReturn:
-    __metaclass__ = PoolMeta
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('done')
+    @process_sale('outgoing_moves')
+    def cancel(cls, shipments):
+        super(ShipmentOut, cls).cancel(shipments)
+
+
+class ShipmentOutReturn(metaclass=PoolMeta):
     __name__ = 'stock.shipment.out.return'
 
     @classmethod
@@ -74,35 +75,6 @@ class ShipmentOutReturn:
                     'by a sale.'),
                 })
 
-    @classmethod
-    def write(cls, *args):
-        pool = Pool()
-        Sale = pool.get('sale.sale')
-        SaleLine = pool.get('sale.line')
-
-        super(ShipmentOutReturn, cls).write(*args)
-
-        actions = iter(args)
-        for shipments, values in zip(actions, actions):
-            if values.get('state') != 'received':
-                continue
-            sales = []
-            move_ids = []
-            for shipment in shipments:
-                move_ids.extend([x.id for x in shipment.incoming_moves])
-
-            with Transaction().set_context(_check_access=False):
-                sale_lines = SaleLine.search([
-                        ('moves', 'in', move_ids),
-                        ])
-                if sale_lines:
-                    for sale_line in sale_lines:
-                        if sale_line.sale not in sales:
-                            sales.append(sale_line.sale)
-
-                    sales = Sale.browse([s.id for s in sales])
-                    Sale.process(sales)
-
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
@@ -116,9 +88,28 @@ class ShipmentOutReturn:
 
         return super(ShipmentOutReturn, cls).draft(shipments)
 
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('received')
+    @process_sale('incoming_moves')
+    def receive(cls, shipments):
+        super(ShipmentOutReturn, cls).receive(shipments)
 
-class Move:
-    __metaclass__ = PoolMeta
+
+def process_sale_move(func):
+    @wraps(func)
+    def wrapper(cls, moves):
+        pool = Pool()
+        Sale = pool.get('sale.sale')
+        with Transaction().set_context(_check_access=False):
+            sales = set(m.sale for m in cls.browse(moves) if m.sale)
+        func(cls, moves)
+        if sales:
+            Sale.__queue__.process(sales)
+    return wrapper
+
+
+class Move(metaclass=PoolMeta):
     __name__ = 'stock.move'
     sale = fields.Function(fields.Many2One('sale.sale', 'Sale', select=True),
         'get_sale', searcher='search_sale')
@@ -128,24 +119,6 @@ class Move:
         ('recreated', 'Recreated'),
         ], 'Exception State'), 'get_sale_exception_state')
 
-    @classmethod
-    def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().connection.cursor()
-        sql_table = cls.__table__()
-
-        super(Move, cls).__register__(module_name)
-
-        table = TableHandler(cls, module_name)
-
-        # Migration from 2.6: remove sale_line
-        if table.column_exist('sale_line'):
-            cursor.execute(*sql_table.update(
-                    columns=[sql_table.origin],
-                    values=[Concat('sale.line,', sql_table.sale_line)],
-                    where=sql_table.sale_line != Null))
-            table.drop_column('sale_line')
-
     @classmethod
     def _get_origin(cls):
         models = super(Move, cls)._get_origin()
@@ -204,35 +177,11 @@ class Move:
     @classmethod
     @ModelView.button
     @Workflow.transition('cancel')
+    @process_sale_move
     def cancel(cls, moves):
-        pool = Pool()
-        Sale = pool.get('sale.sale')
-        SaleLine = pool.get('sale.line')
-
         super(Move, cls).cancel(moves)
 
-        sale_lines = SaleLine.search([
-                ('moves', 'in', [m.id for m in moves]),
-                ])
-        if sale_lines:
-            sale_ids = list(set(l.sale.id for l in sale_lines))
-            sales = Sale.browse(sale_ids)
-            Sale.process(sales)
-
     @classmethod
+    @process_sale_move
     def delete(cls, moves):
-        pool = Pool()
-        Sale = pool.get('sale.sale')
-        SaleLine = pool.get('sale.line')
-
-        with Transaction().set_context(_check_access=False):
-            sale_lines = SaleLine.search([
-                    ('moves', 'in', [m.id for m in moves]),
-                    ])
-
         super(Move, cls).delete(moves)
-
-        if sale_lines:
-            sales = list(set(l.sale for l in sale_lines))
-            with Transaction().set_context(_check_access=False):
-                Sale.process(sales)
